package obfui::axsyms;

use Win32::OLE::Const;
use Win32::OLE::TypeInfo;
use strict;

my @libs;
sub libCLSID    () {0}
sub libNAME     () {1}
sub libMAJOR    () {2}
sub libMINOR    () {3}
sub libLANGUAGE () {4}
sub libFILENAME () {5}

my %param_names;
my $need_param_names;
sub doit
{
    my ($havevals,$needevents,$need_paramnames_v,@libids) = @_;
    if ($havevals)
    {
	#my @libids = @libs[@libidxs];
	%param_names = ();
	$need_param_names = $need_paramnames_v;
	my $rlist = _get_idents($needevents,\@libids);
	my %h;
	foreach (@$rlist) { $h{$_} = 1 };
	foreach (keys %param_names) { $h{$_} = 1; };
	return [sort keys %h];
    } else {
	my @items = map { [$_->[libNAME],
	    $_->[libMAJOR].'.'.$_->[libMINOR] ] } @libs;
	return \@items;
    }
}


sub _loadliblist
{
    Win32::OLE::Const->EnumTypeLibs(sub {
	my ($clsid,$title,$version,$langid,$filename) = @_;
	return unless $version =~ /^([0-9a-fA-F]+)\.([0-9a-fA-F]+)$/;
	my ($maj,$min) = (hex($1), hex($2));
	push @libs, [$clsid,$title,$maj,$min,$langid,$filename];
    });
    @libs = sort {lc $a->[libNAME] cmp lc $b->[libNAME]} @libs
;
}

_loadliblist();



sub _get_idents
{
    my ($needevts,$libids) = @_;
    my @reslist;    
    foreach my $id (@$libids)
    {
	_get_idents_from_lib(\@reslist,$id,$needevts);
    }
    \@reslist;
}

my $ShowHidden = 0;

#global state
my (@Type,@Member);
sub typeLIB    () {0}
sub typeINFO   () {1}
sub typeDOC    () {2}
sub typeATTR   () {3}
sub typeHIDDEN () {4}

sub membTYPE     () {0}
sub membDESC     () {1}
sub membDOC      () {2}
sub membICON     () {3}
sub membREADONLY () {4}
sub membHIDDEN   () {5}

my ($Type,$Member);

sub _get_idents_from_lib
{
    my ($reslist,$id,$needevents)= @_;
    my @def = @{$libs[$id]}[libNAME,libMAJOR,libMINOR,libLANGUAGE];
    $def[0] = quotemeta $def[0];
    my $typelib = Win32::OLE::Const->LoadRegTypeLib(@def);
    return if (Win32::OLE->LastError);
    my $tlib = $typelib;
    my $tcount = $tlib->_GetTypeInfoCount;

    my %hide;
    for (0..$tcount-1) {
	my $tinfo = $tlib->_GetTypeInfo($_);
	foreach (0..$tinfo->_GetTypeAttr->{cImplTypes}-1) {
	    my $impltinfo = $tinfo->_GetImplTypeInfo($_);
	    next unless defined $impltinfo;
	    ++$hide{$impltinfo->_GetTypeAttr->{guid}};
	}
    }
    undef @Type;
    for (0..$tcount-1) {
	my $tinfo = $tlib->_GetTypeInfo($_);
	my $doc  = $tinfo->_GetDocumentation;
	my $attr = $tinfo->_GetTypeAttr;
	my $tflags = $attr->{wTypeFlags};
	next if $tflags & TYPEFLAG_FRESTRICTED;
	next if $hide{$attr->{guid}};
	#next unless $icon[$attr->{typekind}];
	my $hidden = $tflags & TYPEFLAG_FHIDDEN;
	$hidden = 1 if $doc->{Name} =~ /^_/;
	push @Type, [$tlib, $tinfo, $doc, $attr, $hidden];
    }

    my @Index = grep {
 	$ShowHidden || !$Type[$_]->[typeHIDDEN]
  } 0..@Type-1;
    foreach (@Index)
    {
	_build_member_list($reslist,$id,$_,$needevents);
    }
}

sub _build_member_list
{
    my ($reslist,$libid,$id,$needevents)  = @_;
    undef @Member;
    $Type = $id;
    my $tkind = $Type[$Type]->[typeATTR]->{typekind};
    if ($tkind == TKIND_COCLASS) {
	my ($dispatch,$event);
	my $tinfo = $Type[$Type]->[typeINFO];
	for my $impltype (0 .. $Type[$Type]->[typeATTR]->{cImplTypes}-1) {
	    my $tflags = $tinfo->_GetImplTypeFlags($impltype);
	    next unless $tflags & IMPLTYPEFLAG_FDEFAULT;
	    my $impltinfo = $tinfo->_GetImplTypeInfo($impltype);
	    next unless defined $impltinfo;
	    ($tflags & IMPLTYPEFLAG_FSOURCE ? $event : $dispatch) = $impltinfo;
	}
	addFunctions($dispatch);
	addFunctions($event, 'Event');
    }
    else {
	addFunctions($Type[$Type]->[typeINFO]);
	addVariables($Type[$Type]->[typeINFO]);
    }
    undef $Member;
    #now read @Member and harvest symbols

    map {
	push @$reslist, $Member[$_]->[membDOC]->{Name}
    } grep {
	($ShowHidden || !$Member[$_]->[membHIDDEN]) &&
		($Member[$_]->[membICON] eq 'Event' ?
			$needevents : !$needevents)

    } 0..@Member-1;
    
}

sub addFunctions {
    my ($tinfo,$event) = @_;
    return unless defined $tinfo;
    my $attr = $tinfo->_GetTypeAttr;
    my %property;
    for my $func (0 .. $attr->{cFuncs}-1) {
	my $desc = $tinfo->_GetFuncDesc($func);
	next if $desc->{wFuncFlags} & FUNCFLAG_FRESTRICTED;
	my $doc = $tinfo->_GetDocumentation($desc->{memid});
	my $name = $doc->{Name};
	my $invkind = $desc->{invkind};
	next if $event && $invkind != INVOKE_FUNC;

	if ($invkind != INVOKE_FUNC && exists $property{$name}) {
	    if ($invkind & (INVOKE_PROPERTYPUT | INVOKE_PROPERTYPUTREF)) {
		$Member[$property{$name}]->[membREADONLY] = 0;
	    }
	    if ($invkind == INVOKE_PROPERTYGET) { # prefer GET syntax
		$Member[$property{$name}]->[membDESC] = $desc;
	    }
	}
	else {
	    $property{$name} = @Member if $invkind != INVOKE_FUNC;
	    my $icon = $invkind == INVOKE_FUNC ? ($event||'Function') : 'Property';
	    my $readonly = $invkind == INVOKE_PROPERTYGET;
	    my $hidden = $desc->{wFuncFlags} & FUNCFLAG_FHIDDEN;
	    $hidden = 1 if $doc->{Name} =~ /^_/;
	    push @Member, [$tinfo, $desc, $doc, $icon, $readonly, $hidden];
	    if (!$hidden  && $invkind == INVOKE_FUNC && $need_param_names)
	    {
		my $cParams = $desc->{cParams};
		my $names = $tinfo->_GetNames($desc->{memid}, $cParams+1);
		for my $param (0 .. $cParams-1) {
		    my $nm = $names->[$param];
		    $param_names{$nm} = 1;
		};
	    };
	}
    }
}

sub addVariables {
    my $tinfo = shift;
    return unless defined $tinfo;
    my $attr = $tinfo->_GetTypeAttr;
    for my $var (0 .. $attr->{cVars}-1) {
	my $desc = $tinfo->_GetVarDesc($var);
	next if $desc->{wVarFlags} & VARFLAG_FRESTRICTED;
	my $doc = $tinfo->_GetDocumentation($desc->{memid});
	push @Member, [$tinfo, $desc, $doc, 'Const'];
    }
}

1;